import os
import json
from json.decoder import JSONDecodeError
import logging
from shapely import wkt
from shapely.geometry import shape, GeometryCollection
from shapely.geometry.base import BaseGeometry
from shapely.errors import WKTReadingError

from max_ard.exceptions import UnknownFileType, MissingDependency

from max_ard.dependency_support import HAS_FIONA
if HAS_FIONA:
    import fiona
    from fiona.collection import Collection as FionaCollection


def convert_to_shapely(input):
    ''' Convert input geometry or geometry filepath to Shapely shape object

        Args:
            input: a geometry object with one or more features, or filepath to such a geometry object, as:
                - a dict that follows the Python Geometry Interface (GeoJSON-like)
                - a GeoJSON feature-like dict that has a 'geometry' key
                - a Fiona dataset reader
                - a Shapely shape *or iterable of Shapely shapes* (in which case the input will be returned as is)

        Returns:
            Shapely shape object of input (or iterable of Shapely shapes, if that was the input)
            
        * Note: if the input has multiple features, the output will be a GeometryCollection *
        '''

    # ignore Nones, empty strings, other falsey things

    if not input:
        return input

    # check if we already have a Shapely shape
    if isinstance(input, BaseGeometry):
        return input

    # or an iterable of Shapely shapes
    try:
        if all([isinstance(item, BaseGeometry) for item in input]):
            return input
    except TypeError:
        pass


    # GeoJSON-like (__geo_interface__) -- single feature
    try:
        return shape(input)
    except:
        pass
    
    try:
        return shape(input['geometry'])
    except:
      pass

    # GeoJSON feature collection 
    try:
        return GeometryCollection([shape(g['geometry']) for g in input['features']])
    except:
        pass

    def fiona_converter(reader):
        geom = list(reader)
        if len(geom) > 1:
            return GeometryCollection([shape(g['geometry']) for g in geom])
        else:
            return shape(geom[0]['geometry']) 

    if HAS_FIONA and type(input) == FionaCollection:
        try: 
            return fiona_converter(input)
        except:
            pass

    if type(input) is not str:
        raise TypeError("Invalid object - check list of acceptable types.")

    def quiet_wkt_loads(input):
        logger = logging.getLogger('shapely.geos')
        old_level = logger.level
        logger.setLevel(logging.CRITICAL)
        try:
            geom = wkt.loads(input)
        except Exception as e:
            raise(e)
        finally:
            logger.setLevel(old_level)
        return geom

    # strings could be wkt, json, or a path to a vector file
    if os.path.exists(input):
        try:
             return quiet_wkt_loads(open(input, 'r').read())
        # try to read filepath with Fiona
        except (WKTReadingError, UnicodeDecodeError):
            if not HAS_FIONA:
                raise MissingDependency('Fiona is needed to read vector formats')
            try:
                reader = fiona.open(input)
                return fiona_converter(reader)
            except:
                raise UnknownFileType('Could not open the contents of the given file, is it a valid file?')

    else:

        # wkt string 
        try:
            return quiet_wkt_loads(input)
        except WKTReadingError:
            pass

        # json string
        try:
            return shape(json.loads(input))
        except JSONDecodeError:
            pass

        raise TypeError('String input does not match WKT or JSON, or a valid file path')

