"""Tools and utilities for processing ARD images

Provides
--------
- read/write windows from ARD collections
"""

from typing import Any
from shapely.ops import transform

from maxar_canvas_grid.grid import get_transform, get_CRS
from maxar_canvas_grid import covers

from max_ard.io import convert_to_shapely

from max_ard.dependency_support import HAS_RASTERIO
from max_ard.exceptions import MissingDependency

if HAS_RASTERIO:
    import rasterio
    from rasterio import windows


        
__all__ = ('iterify', 'read_windows', 'write_windows')

class COGReader():
    ''' An optimized COG rasterio reader

    ARD COGs do not have any sidecar files, therefore when opening
    we can use GDAL_DISABLE_READDIR_ON_OPEN='EMPTY_DIR' to skip looking
    for sidecar files. This skips an S3 LIST operation which takes
    time and money (LISTs cost more than READs).

    We also override some GDAL read settings for more potential speed
    and cost gains

    Parameters
    ----------
    path : str
        Path of COG to open
    *args : list
        Any additional args to pass to rasterio.open()
    **kwargs : dict
        Any additional keyword arguments to pass to rasterio.open() '''

    COG_ENV = {
        'GDAL_HTTP_MERGE_CONSECUTIVE_RANGES': True,
        'VSI_CACHE': True,
        'GDAL_HTTP_MULTIPLEX': True
    }

    def __init__(self, path, *args, **kwargs):
        with rasterio.Env(GDAL_DISABLE_READDIR_ON_OPEN='EMPTY_DIR'):
            self.dsr = rasterio.open(path, *args, **kwargs)

    def __getattribute__(self, name: str) -> Any:
        ''' Proxy to the underlying rasterio Dataset Reader '''
        if name == 'read':
            def func(*args, **kwargs):
                with rasterio.Env(**COGReader.COG_ENV):
                    return self.dsr.read(*args, **kwargs)
            return func
        else:
            dsr = object.__getattribute__(self, 'dsr')
            if name == 'dsr':
                return dsr
            else:
                return getattr(dsr, name)

    # context manager dunders can't be proxied
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.dsr.close() 


def iterify(input):
    '''Takes an input of any type and returns a list containing that input, if the input is not already an iterable or if it is a string.
    If the input is a non-string iterable, returns the input as is.'''
    
    if isinstance(input, str):
        return [input]
    else:
        try:
            input[0]
            return input
        except:
            return [input]

def loop_geoms(collection, geoms, src_proj=4326):
    ''' Loops over a geometry source and emits a tuple of:
        - An ARDTile that overlaps a feature from the source
        - The feature as a Shapely geometry, reprojected to the tile's UTM projection

        Args:
            geoms: one or more geometry objects as:
                - Shapely geometry objects
                - dicts that follow the Python Geometry Interface (GeoJSON-like)
                - GeoJSON feature-like dicts that have a 'geometry' key
                - lists or tuples of above
                - a Fiona dataset reader
                - a path to a readable geometry file.

        Yields:
            tuple(ARDTile, shape): tuple of ARD Tile and reprojected Shapely geometry object'''

    src_crs = get_CRS(src_proj)

    geoms = convert_to_shapely(geoms)
    for geom in iterify(geoms):
        for cell in covers(geom, src_proj=src_proj):
            for tile in collection.get_stack(cell):
                tile_crs = get_CRS(tile.cell.proj)
                tfm = get_transform(src_crs, tile_crs)
                proj_geom = transform(tfm, geom)
                yield tile, proj_geom


def read_windows(collection, geoms, src_proj=4326):
    ''' A generator for reading windows from overlapping tiles in an ARDCollection

            Args:
                geoms: one or more geometry objects as:
                    - Shapely geometry objects
                    - dicts that follow the Python Geometry Interface (GeoJSON-like)
                    - GeoJSON feature-like dicts that have a 'geometry' key
                    - lists or tuples of above
                    - a Fiona dataset reader
                src_proj: An EPSG identifier parseable by Proj4

            Yields:
                Tuple of:
                    ARDTile of tile
                    Shapely polygon representing the bounds of the input geometry
                        in the tile's UTM coordinate system
                    a Reader function:

                        reader(asset)

                            Args:
                                asset(str): asset to read: `visual`, `pan`, or `ms`
                            
                            Returns:
                                numpy array of data covered by the returned geometry

            Example:

            geoms = [...]
            for tile, geom, reader in read_windows(geoms):
                # can check tile.properties, or fetch tile.clouds
                # can check if `geom` intersects with clouds
                data = reader('visual') '''

    if not HAS_RASTERIO:
        raise MissingDependency('Rasterio is required to read ARD rasters')

    for tile, geom in loop_geoms(collection, geoms, src_proj):

        def reader(asset, **kwargs):
            dsr = tile.open_asset(asset)
            window = rasterio.windows.from_bounds(*geom.bounds, dsr.transform)
            return dsr.read(window=window, **kwargs)

        yield (tile, geom, reader)


def write_windows(collection, geoms, src_proj=4326):

    if not HAS_RASTERIO:
        raise MissingDependency('Rasterio is required to read ARD rasters')

    for tile, geom in loop_geoms(collection, geoms, src_proj):
    
        def writer(asset, name, **kwargs):
            dsr = tile.open_asset(asset)
            window = rasterio.windows.from_bounds(*geom.bounds, dsr.transform)
            window = windows.from_bounds(*geom.bounds, dsr.transform)
            profile = dsr.meta
            profile.update({
                'width': window.width,
                'height': window.height,
                'count': dsr.count,
                'transform': dsr.window_transform(window),
                'driver': 'GTiff'
            })
            profile.update(kwargs)
            with rasterio.open(name, 'w', **profile) as dest:
                dest.write(dsr.read(window=window))

        yield (tile, geom, writer)